/*****************************************************************************
 * Includes
******************************************************************************/

/*
 * System includes
 */
#include <Application.h>
#include <Bitmap.h>
#include <File.h>
#include <FindDirectory.h>
#include <Node.h>
#include <OS.h>
#include <Screen.h>
#include <ScrollView.h>
#include <View.h>
#include <Window.h>

#include <stdio.h>

/*
 * POV specific includes
 */
#include "frame.h"
#include "parse.h"     /* For MAError */
#include "povproto.h"
#include "povray.h"

/*****************************************************************************
* Local preprocessor defines
******************************************************************************/

/*****************************************************************************
* Local typedefs
******************************************************************************/
#define POV_SIGNATURE  "application/x-vnd.pov-ray"
#define POV_WIN_TITLE  "Persistence Of Vision"
#define POV_VIEW_NAME  "POV View"
#define POV_APP_THREAD "POV BApp Thread"
#define POV_PREFS_FILE "povray.prefs"

/*****************************************************************************
* Globally-Local variables
******************************************************************************/
BApplication* povApp;
BWindow*      povWindow;
BView*        povView;
BBitmap*      povBitmap;
BView*        backView;
thread_id     appThread;
sem_id        endSemaphore;
bool          displayActive = false;
/* Quite awful */
int32  imgWidth;
int32  imgHeight;
BPoint winOrigin;

/*****************************************************************************
* Static functions
******************************************************************************/
void ReadPreferences()
{
	ssize_t  rs;
	char     path[B_PATH_NAME_LENGTH];
	
	// default values
	winOrigin = BPoint(100,100);
	
	// read prefs file
	find_directory(B_USER_SETTINGS_DIRECTORY, 0, false, path, 256);
	strcat(path,"/"); 
	strcat(path,POV_PREFS_FILE);
	BNode prefNode(path);
	if (prefNode.InitCheck() != B_OK) {
		return; // Default values will be used
	}		
	
	rs = prefNode.ReadAttr("WinPosition", B_POINT_TYPE, 0, &winOrigin, sizeof(BPoint));
	if (rs < (ssize_t)sizeof(BPoint))
		winOrigin = BPoint(100,100);
}

void WritePreferences()
{
	ssize_t  rs;
	char     path[B_PATH_NAME_LENGTH];
	status_t rc;
		
	// write prefs file
	find_directory(B_USER_SETTINGS_DIRECTORY, 0, false, path, 256);
	strcat(path,"/"); 
	strcat(path,POV_PREFS_FILE);
	
	BFile prefFile;
	rc = prefFile.SetTo(path, B_WRITE_ONLY | B_CREATE_FILE);
	if (rc != B_OK) {
		return;
	}

	BNode prefNode(path);
	if (prefNode.InitCheck() != B_OK) {
		return;
	}
		
	rs = prefNode.WriteAttr("WinPosition", B_POINT_TYPE, 0, &winOrigin, sizeof(BPoint));		
}

/*****************************************************************************
* Local classes
******************************************************************************/
class PovWin : public BWindow {
	public:
		PovWin(BRect frame);
		virtual bool QuitRequested();
		virtual void FrameMoved(BPoint origin);
};

PovWin::PovWin(BRect frame) 
	: BWindow(frame, POV_WIN_TITLE, B_DOCUMENT_WINDOW, 
	          B_NOT_ZOOMABLE /*| B_NOT_CLOSABLE */)
{
}

bool PovWin::QuitRequested()
{
	displayActive = false;
	release_sem(endSemaphore);
	return true;
}

void PovWin::FrameMoved(BPoint origin)
{
	winOrigin = origin;
	WritePreferences();
}

class PovView : public BView {
public:
	PovView(BRect frame);
	virtual void AttachedToWindow(void);
	virtual void Draw(BRect updateRect);
	virtual void FrameResized(float w, float h);
	virtual void KeyDown(const char* bytes, int32 numBytes);
	virtual void MouseDown(BPoint point);
	void FixupScrollbars(void);
};

PovView::PovView(BRect frame) 
	: BView(frame, POV_VIEW_NAME, B_FOLLOW_ALL_SIDES,
	        B_WILL_DRAW | B_FRAME_EVENTS)
{
}

void PovView::Draw(BRect updateRect) 
{
	if (povBitmap != NULL)
	{
		DrawBitmap(povBitmap,updateRect,updateRect);
		Sync();
	}
}

void PovView::FrameResized(float w, float h) 
{
	FixupScrollbars();
}

void PovView :: AttachedToWindow(void)
{
	FixupScrollbars();
	SetViewColor(B_TRANSPARENT_32_BIT);
	MakeFocus(); /* thus the keyboard events are received */
}

// Copied from Be sample-code scrollbarapp
void PovView::FixupScrollbars(void)
{
	BRect bounds;
	BScrollBar *sb;

	bounds = Bounds();
	float myPixelWidth = bounds.Width();
	float myPixelHeight = bounds.Height();
	// [BT] peut-etre erreur a un pret
	float maxWidth = imgWidth-1;
	float maxHeight = imgHeight-1;
		
	float propW = myPixelWidth/maxWidth;
	float propH = myPixelHeight/maxHeight;
	
	float rangeW = maxWidth - myPixelWidth;
	float rangeH = maxHeight - myPixelHeight;
	if(rangeW < 0) rangeW = 0;
	if(rangeH < 0) rangeH = 0;

	if ((sb=ScrollBar(B_HORIZONTAL))!=NULL) {
		sb->SetProportion(propW);
		sb->SetRange(0,rangeW);
		// Steps are 1/8 visible window for small steps
		//   and 1/2 visible window for large steps
		sb->SetSteps(myPixelWidth / 8.0, myPixelWidth / 2.0);
	} 

	if ((sb=ScrollBar(B_VERTICAL))!=NULL) {
		sb->SetProportion(propH);
		sb->SetRange(0,rangeH);
		// Steps are 1/8 visible window for small steps
		//   and 1/2 visible window for large steps
		sb->SetSteps(myPixelHeight / 8.0, myPixelHeight / 2.0);
	}
}

void PovView::KeyDown(const char* bytes, int32 numBytes) 
{
	if ( numBytes == 1 ) {
		switch ( bytes[0] ) {
			/* the user has  pressed "Q", let's release the end semaphore*/
			case 'q':
			case 'Q':
				if (displayActive) {
					release_sem(endSemaphore);
				}
				break;
		}
	}
}

void PovView::MouseDown(BPoint point)
{
	if (displayActive) {
		/* the user has clicked in the view, let's release the end semaphore*/
		release_sem(endSemaphore);
	}
}

/*****************************************************************************
*
* FUNCTION  BEOS_init_povray
*
* INPUT     argc, argv - the command-line arguments
*
* OUTPUT
*
* RETURNS   argc, argv - command-line args with the BeOS specific
*                        arguments removed and/or converted into something
*                        corresponding to a POV-Ray option
*
* AUTHOR    B.Thery
*
* DESCRIPTION Create a BApplication. This is mandatory if we want to connect
*             to the app_server and create windows.
*
* CHANGES
*
*
******************************************************************************/
void BEOS_init_povray(int *argc, char** argv[])
{
	displayActive = false;

	povApp = new BApplication(POV_SIGNATURE);

	/* Create semaphore used at the end of the rendering */
	endSemaphore = create_sem(0, "wait user");
}

/*****************************************************************************
*
* FUNCTION   BEOS_finish_povray
*
* INPUT      exit status
*
* OUTPUT
*
* RETURNS
*
* AUTHOR       B.Thery
*
* DESCRIPTION  Deletes the be BApplication created in BEOS_init_povray()
*
* CHANGES
*
******************************************************************************/
void BEOS_finish_povray(int status)
{
// Release 3 : Moved to BEOS_display_finished() 
//	if (displayActive) {
//		povWindow->Lock();
//		povWindow->Quit();
//		delete povBitmap;
//	}
	/* Being brutal avoid complaints */
	kill_thread(appThread);
	delete povApp;
	exit(status);	
}

/*****************************************************************************
*
* FUNCTION  BEOS_display_init
*
* INPUT  width, height - size of desired window
*
* OUTPUT
*
* RETURNS
*    global variables povWindow, povView set up
*
* AUTHOR    B.Thery
*
* DESCRIPTION
*
*   Create the window. Sets the window properties.
*
* CHANGES
*
******************************************************************************/
void BEOS_display_init(int width, int height)
{
	imgWidth = width;
	imgHeight = height;
	BScrollView* scrollView;	
	
	displayActive = true;
	
	ReadPreferences();
	
	/* [BT] Customize, remember position */
	BRect rect(winOrigin.x,winOrigin.y, 
	           winOrigin.x+(width-1)+ B_V_SCROLL_BAR_WIDTH,
               winOrigin.y+(height-1)+ B_H_SCROLL_BAR_HEIGHT);
               
    /* Create window */
	povWindow = new PovWin(rect);
	povWindow->SetSizeLimits(32+ B_V_SCROLL_BAR_WIDTH,(width-1)+ B_V_SCROLL_BAR_WIDTH,
							 32+ B_H_SCROLL_BAR_HEIGHT,(height-1)+ B_H_SCROLL_BAR_HEIGHT);

	/* Create display view with scrollbars */	 
	rect = BRect(0,0,width-1, height-1);
	
	povView = new PovView(rect);
    povView->SetViewColor(255,0,0);
	
	scrollView = new BScrollView("POV scroll view", povView, B_FOLLOW_ALL_SIDES,
								0, true, true, B_NO_BORDER);
	povWindow->AddChild(scrollView);
	
	/* Create offscreen bitmap */								 
	povBitmap = new BBitmap(rect, B_RGBA32, true);
	backView = new BView(rect, POV_VIEW_NAME, B_FOLLOW_NONE, 0);
	povBitmap->AddChild(backView);
	
	/* Dirty update of size */
	BScreen screen(povWindow);
	povWindow->ResizeTo(screen.Frame().Width() - winOrigin.x - B_V_SCROLL_BAR_WIDTH,
	                    screen.Frame().Height() - winOrigin.y - B_H_SCROLL_BAR_HEIGHT);

	povWindow->Show();
}

/*****************************************************************************
*
* FUNCTION   BEOS_display_plot
*
* INPUT   x, y location of pixel
*         Red, Green, Blue, Alpha color components of pixel
*
* OUTPUT
*
* RETURNS
*
* AUTHOR    B.Thery
*
* DESCRIPTION
*
*    It is too slow to write each pixel to the screen as it is generated. For
*    simple pictures, we spend far too much time drawing on the screen instead
*    of raytracing. So, we'll update the window with the contents of the image
*    when we reach the end of the scanline, or if we have been rendering for
*    several seconds without an update.
*
* CHANGES
*
******************************************************************************/
void BEOS_display_plot(int x, int y, unsigned char Red, unsigned char Green,
                       unsigned char Blue, unsigned char Alpha)
{
	rgb_color c; 
	c.red = Red; c.green = Green; c.blue = Blue; c.alpha = Alpha;

	if (!displayActive)
		return;

	povBitmap->Lock();
		backView->SetHighColor(c);
		backView->StrokeLine(BPoint(x,y),BPoint(x,y));
		if (x == imgWidth-1) {
			backView->Sync();
			povView->LockLooper();
				povView->Invalidate(BRect(0,y,x,y));
			if (displayActive) // to avoid error when closing window
				povView->UnlockLooper();
		}
	povBitmap->Unlock();
}

/*****************************************************************************
*
* FUNCTION   BEOS_display_plot_rect
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR  B.Thery
*
* DESCRIPTION
*
*    It is too slow to write each pixel to the screen as it is generated. For
*    simple pictures, we spend far too much time drawing on the screen instead
*    of raytracing. So, we'll update the window with the contents of the image
*    when we reach the end of the scanline, if we have just output a large
*    block, or if we have been rendering for several seconds without an update,
*    but only if we can see it.  We call the PlotPixel routine for each pixel
*    rather than just drawing a block, as this allows us to dither the large
*    blocks if we don't have the exact color.
*
*    We reduce the dithering errors to the left and the right of the block
*    just displayed because the errors accumulate there for each row in the
*    block, but they only get "used" by the next block to the right or the
*    block to the lower left.  By dividing the errors by the height of the
*    block just displayed, we "average" the error from the previous block
*    for the edge of the next block.
*
* CHANGES
*
*
******************************************************************************/
void BEOS_display_plot_rect(int x1, int y1, int x2, int y2, unsigned char Red,
                unsigned char Green, unsigned char Blue, unsigned char Alpha)
{
	rgb_color c; 
	c.red = Red; c.green = Green; c.blue = Blue; c.alpha = Alpha;

	if (!displayActive)
		return;

	povBitmap->Lock();
		backView->SetHighColor(c);
		backView->FillRect(BRect(x1,y1,x2,y2));
		backView->Sync();
	povBitmap->Unlock();	
}

/*****************************************************************************
*
* FUNCTION   BEOS_display_plot_box
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR     B.Thery
*
* DESCRIPTION
*
*    Draw a rectangular hollow box on the screen.  
*
* CHANGES 
*
******************************************************************************/
void BEOS_display_plot_box(int x1, int y1, int x2, int y2, unsigned char Red,
                unsigned char Green, unsigned char Blue, unsigned char Alpha)
{
	rgb_color c; 
	c.red = Red; c.green = Green; c.blue = Blue; c.alpha = Alpha;

	if (!displayActive)
		return;

	povBitmap->Lock();
		backView->SetHighColor(c);
		backView->StrokeRect(BRect(x1,y1,x2,y2));
		backView->Sync();
	povBitmap->Unlock();		
}

/*****************************************************************************
*
* FUNCTION  BEOS_display_finished
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR 		B.Thery
*
* DESCRIPTION
*
*   If option PAUSE WHEN DONE (+P) is set :
*   Wait for user to press a mouse button or the 'q' key in the rendering
*   window, then return to the caller.
*
* CHANGES
*
******************************************************************************/
void BEOS_display_finished()
{
	if (displayActive && (opts.Options & PROMPTEXIT)) {
		/* Wait for the semaphore to be incremented by a user action */
		/* see PovView::MouseDown() or PovView::KeyDown() */
		acquire_sem(endSemaphore);
	}
	if (displayActive) {
		povWindow->Lock();
		povWindow->Quit();
		delete povBitmap;
	}
}

/*****************************************************************************
*
* FUNCTION  BEOS_Test_Abort
*
* INPUT
*
* OUTPUT
*
* RETURNS  whether or not the user wants to quit
*
* AUTHOR   B.Thery
*
* DESCRIPTION
*
*   This is the polling user abort checker called by POV-Ray for TEST_ABORT
*   if the EXITENABLE flag is set (+x).  This is called every pixel, or less
*   often if the +Xnnn switch is set to test after every nnn pixels.
*
* CHANGES
*
******************************************************************************/
int BEOS_Test_Abort()
{
	//[BT] TO BE IMPLEMENTED
	return FALSE;
}

/*****************************************************************************
*
* FUNCTION  BEOS_display_close
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*    Reset the values used by the X display, in case we are animating.
*
* CHANGES
*
*    Moved memory freeing and X display closing to XWIN_finish_povray
*    so that we dont re-do everything during an animation.  July, 1996 [AED]
*
******************************************************************************/
void BEOS_display_close()
{
	//[BT] TO BE IMPLEMENTED
	povWindow->Close();
}
